"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fibers = require('fibers');
///
/// ## run/wait
/// * `promise = run(() => { wait(promise/callback); ... })`
///    create a coroutine to write asynchronous code in a synchronous way.
///
/// All the job is done by fibers library.
/// Those two functions are the core, others are goodies:
///   * `promise = run(fn)` create a coroutine.
///     This start a fiber, which is stopped when `fn` returns.
///   * `result = wait(promise/callback)` encapsulate promise or callback.
///     Concretely, the fiber is suspended while the asynchronous task is not finished, then it resumes.
///     As many `wait()` as needed may be used in a run.
exports.wait = (promiseOrCallback) => {
    const fiber = fibers.current;
    if (!fiber)
        throw new Error('cannot wait: no fiber');
    if (typeof promiseOrCallback === 'function') {
        promiseOrCallback((err, res) => {
            process.nextTick(() => {
                let cx = globals.context;
                try {
                    if (err) {
                        fiber.throwInto(err);
                    }
                    else {
                        fiber.run(res);
                    }
                }
                finally {
                    globals.context = cx;
                    cx = null;
                }
            });
        });
    }
    else {
        promiseOrCallback
            .then(res => {
            let cx = globals.context;
            try {
                fiber.run(res);
            }
            finally {
                globals.context = cx;
                cx = null;
            }
        })
            .catch(e => {
            let cx = globals.context;
            try {
                fiber.throwInto(e);
            }
            finally {
                globals.context = cx;
                cx = null;
            }
        });
    }
    let cx = globals.context;
    try {
        return fibers.yield();
    }
    catch (e) {
        throw (fullStackError && fullStackError(e)) || e;
    }
    finally {
        globals.context = cx;
        cx = null;
    }
};
exports.run = (fn) => {
    if (typeof fn !== 'function') {
        throw new Error('run() should take a function as argument');
    }
    return new Promise((resolve, reject) => {
        const cx = globals.context;
        fibers(() => {
            try {
                resolve(fn());
            }
            catch (e) {
                reject((cleanFiberStack && cleanFiberStack(e)) || e);
            }
        }).run();
        globals.context = cx;
    });
};
// goodies
/// ## funnel
/// * `fun = funnel(max)`
///   limits the number of concurrent executions of a given code block.
///
/// The `funnel` function is typically used with the following pattern:
///
/// ``` ts
/// // somewhere
/// var myFunnel = funnel(10); // create a funnel that only allows 10 concurrent executions.
///
/// // elsewhere
/// myFunnel(function() { /* code with at most 10 concurrent executions */ });
/// ```
///
/// The `funnel` function can also be used to implement critical sections. Just set funnel's `max` parameter to 1.
///
/// If `max` is set to 0, a default number of parallel executions is allowed.
/// This default number can be read and set via `flows.funnel.defaultSize`.
/// If `max` is negative, the funnel does not limit the level of parallelism.
///
/// The funnel can be closed with `fun.close()`.
/// When a funnel is closed, the operations that are still in the funnel will continue but their callbacks
/// won't be called, and no other operation will enter the funnel.
function funnel(max = -1) {
    if (typeof max !== 'number') {
        throw new Error('bad max number: ' + max);
    }
    const _max = max === 0 ? exports.funnel.defaultSize : max;
    // Each bottled coroutine use an handshake to be waked up later when an other quit.
    // Before waiting on the handshake, it is pushed to this queue.
    let queue = [];
    let active = 0;
    let closed = false;
    function tryEnter(fn) {
        if (active < _max) {
            active++;
            try {
                return fn();
            }
            finally {
                active--;
                const hk = queue.shift();
                if (hk) {
                    hk.notify();
                }
            }
        }
        else {
            return overflow(fn);
        }
    }
    function overflow(fn) {
        const hk = handshake();
        queue.push(hk);
        hk.wait();
        if (closed) {
            throw new Error(`cannot execute: funnel has been closed`);
        }
        // A success is not sure, the entry ticket may have already be taken by another,
        // so this one may still be delayed by re-entering in overflow().
        return tryEnter(fn);
    }
    const fun = function (fn) {
        if (closed) {
            throw new Error(`cannot execute: funnel has been closed`);
        }
        if (_max < 0 || _max === Infinity) {
            return fn();
        }
        return tryEnter(fn);
    };
    fun.close = () => {
        queue.forEach(hk => {
            hk.notify();
        });
        queue = [];
        closed = true;
    };
    return fun;
}
exports.funnel = funnel;
funnel.defaultSize = 4;
///
/// ## handshake and queue
/// * `hs = handshake()`
///   allocates a simple semaphore that can be used to do simple handshakes between two tasks.
///   The returned handshake object has two methods:
///   `hs.wait()`: waits until `hs` is notified.
///   `hs.notify()`: notifies `hs`.
///   Note: `wait` calls are not queued. An exception is thrown if wait is called while another `wait` is pending.
function handshake() {
    let callback = undefined, notified = false;
    return {
        wait() {
            return exports.wait((cb) => {
                if (callback)
                    throw new Error('already waiting');
                if (notified)
                    setImmediate(cb);
                else
                    callback = cb;
                notified = false;
            });
        },
        notify() {
            if (!callback)
                notified = true;
            else
                setImmediate(callback);
            callback = undefined;
        },
    };
}
exports.handshake = handshake;
class Queue {
    constructor(options) {
        this._q = [];
        this._pendingWrites = [];
        if (typeof options === 'number') {
            options = {
                max: options,
            };
        }
        options = options || {};
        this._max = options.max != null ? options.max : -1;
    }
    read() {
        return exports.wait((cb) => {
            if (this._callback)
                throw new Error('already getting');
            if (this._q.length > 0) {
                const item = this._q.shift();
                // recycle queue when empty to avoid maintaining arrays that have grown large and shrunk
                if (this._q.length === 0)
                    this._q = [];
                setImmediate(() => {
                    cb(this._err, item);
                });
                if (this._pendingWrites.length > 0) {
                    const wr = this._pendingWrites.shift();
                    setImmediate(() => {
                        wr && wr[0](this._err, wr[1]);
                    });
                }
            }
            else {
                this._callback = cb;
            }
        });
    }
    ///   `q.write(data)`:  queues an item. Waits if the queue is full.
    write(item) {
        return exports.wait((cb) => {
            if (this.put(item)) {
                setImmediate(() => {
                    cb(this._err);
                });
            }
            else {
                this._pendingWrites.push([cb, item]);
            }
        });
    }
    ///   `ok = q.put(data)`: queues an item synchronously. Returns true if the queue accepted it, false otherwise.
    put(item, force) {
        if (!this._callback) {
            if (this._max >= 0 && this._q.length >= this._max && !force)
                return false;
            this._q.push(item);
        }
        else {
            const cb = this._callback;
            this._callback = undefined;
            setImmediate(() => {
                cb(this._err, item);
            });
        }
        return true;
    }
    ///   `q.end()`: ends the queue. This is the synchronous equivalent of `q.write(_, undefined)`
    end() {
        this.put(undefined, true);
    }
    ///   `data = q.peek()`: returns the first item, without dequeuing it. Returns `undefined` if the queue is empty.
    peek() {
        return this._q[0];
    }
    ///   `array = q.contents()`: returns a copy of the queue's contents.
    contents() {
        return this._q.slice(0);
    }
    ///   `q.adjust(fn[, thisObj])`: adjusts the contents of the queue by calling `newContents = fn(oldContents)`.
    adjust(fn) {
        const nq = fn.call(null, this._q);
        if (!Array.isArray(nq))
            throw new Error('adjust function does not return array');
        this._q = nq;
    }
    get length() {
        return this._q.length;
    }
}
exports.Queue = Queue;
///
/// ## Continuation local storage (CLS)
///
/// * `result = withContext(fn, cx)`
///   wraps a function so that it executes with context `cx` (or a wrapper around current context if `cx` is falsy).
///   The previous context will be restored when the function returns (or throws).
///   returns the wrapped function.
function withContext(fn, cx) {
    if (!fibers.current)
        throw new Error('withContext(fn) not allowed outside run()');
    const oldContext = globals.context;
    globals.context = cx || Object.create(oldContext);
    try {
        return fn();
    }
    finally {
        globals.context = oldContext;
    }
}
exports.withContext = withContext;
function context() {
    return globals.context;
}
exports.context = context;
///
/// ## Miscellaneous
///
/// * `results = map(collection, fn)`
///   creates as many coroutines with `fn` as items in `collection` and wait for them to finish to return result array.
function map(collection, fn) {
    return exports.wait(Promise.all(collection.map(item => {
        return exports.run(() => fn(item));
    })));
}
exports.map = map;
/// * `sleep(ms)`
///   suspends current coroutine for `ms` milliseconds.
function sleep(n) {
    exports.wait(cb => setTimeout(cb, n));
}
exports.sleep = sleep;
/// * `ok = canWait()`
///   returns whether `wait` calls are allowed (whether we are called from a `run`).
function canWait() {
    return !!fibers.current;
}
exports.canWait = canWait;
/// * `wrapped = eventHandler(handler)`
///   wraps `handler` so that it can call `wait`.
///   the wrapped handler will execute on the current fiber if canWait() is true.
///   otherwise it will be `run` on a new fiber (without waiting for its completion)
function eventHandler(handler) {
    const wrapped = function (...args) {
        if (canWait()) {
            handler.apply(this, args);
        }
        else {
            exports.run(() => withContext(() => handler.apply(this, args), {})).catch(err => {
                console.error(err);
            });
        }
    };
    // preserve arity
    Object.defineProperty(wrapped, 'length', { value: handler.length });
    return wrapped;
}
exports.eventHandler = eventHandler;
const secret = '_20c7abceb95c4eb88b7ca1895b1170d1';
const globals = (global[secret] = global[secret] || { context: {} });
// Those functions are conditionally assigned bellow.
let fullStackError;
let cleanFiberStack;
let cannotOverrideStackWarned = false;
function overrideStack(e, getFn) {
    try {
        Object.defineProperty(e, 'stack', {
            get: getFn,
            enumerable: true,
            configurable: true,
        });
    }
    catch (e) {
        if (!cannotOverrideStackWarned) {
            console.warn(`[F-PROMISE] attempt to override e.stack failed (warning will not be repeated)`);
            cannotOverrideStackWarned = true;
        }
    }
}
/// ## Error stack traces
///
/// Three policies:
/// * `fast`: stack traces are not changed. Call history might be difficult to read; cost less.
/// * `whole`: stack traces due to async tasks errors in `wait()` are concatenate with the current coroutine stack.
///   This allow to have a complete history call (including f-promise traces).
/// * default: stack traces are like `whole` policy, but clean up to remove f-promise noise.
///
/// The policy can be set with `FPROMISE_STACK_TRACES` environment variable.
/// Any value other than `fast` and `whole` are consider as default policy.
if (process.env.FPROMISE_STACK_TRACES === 'whole') {
    fullStackError = function fullStackError(e) {
        if (!(e instanceof Error)) {
            return e;
        }
        const localError = new Error('__fpromise');
        const fiberStack = e.stack || '';
        overrideStack(e, function () {
            const localStack = localError ? localError.stack || '' : '';
            return fiberStack + localStack;
        });
        return e;
    };
}
else if (process.env.FPROMISE_STACK_TRACES !== 'fast') {
    fullStackError = function fullStackError(e) {
        if (!(e instanceof Error)) {
            return e;
        }
        const localError = new Error('__f-promise');
        const fiberStack = e.stack || '';
        overrideStack(e, function () {
            const localStack = localError ? localError.stack || '' : '';
            return (fiberStack +
                '\n' +
                localStack
                    .split('\n')
                    .slice(1)
                    .filter(line => {
                    return !/\/f-promise\//.test(line);
                })
                    .join('\n'));
        });
        return e;
    };
    cleanFiberStack = function cleanFiberStack(e) {
        if (!(e instanceof Error)) {
            return e;
        }
        const fiberStack = e.stack || '';
        overrideStack(e, function () {
            return fiberStack
                .split('\n')
                .filter(line => {
                return !/\/f-promise\//.test(line);
            })
                .join('\n');
        });
        return e;
    };
}
// little goodie to improve V8 debugger experience
// The debugger hangs if Fiber.yield is called when evaluating expressions at a breakpoint
// So we monkey patch wait to throw an exception if it detects this special situation.
if (process.execArgv.find(str => str.startsWith('--inspect-brk'))) {
    // Unfortunately, there is no public API to check if we are called from a breakpoint.
    // There is a C++ API (context->IsDebugEvaluateContext()) to test this
    // but unfortunately this is an internal V8 API.
    // This test is the best workaround I have found.
    const isDebugEval = () => (new Error().stack || '').indexOf('.remoteFunction (<anonymous>') >= 0;
    const oldWait = exports.wait;
    const flushDelayed = () => {
        if (fibers.current.delayed) {
            const delayed = fibers.current.delayed;
            fibers.current.delayed = undefined;
            for (const arg of delayed) {
                try {
                    oldWait(arg);
                }
                catch (err) {
                    console.error(`delayed 'wait' call failed: ${err.message}`);
                }
            }
        }
    };
    // Why throw string (from bjouhier)
    // I think I threw a string rather than an Error object because
    // I did not want to clutter the debugger with error objects.
    // There was also a memory issue here: debugger allocating a lot
    // of Error objects and stack trace captures vs. a string literal
    // which does not require any dynamic memory allocation.
    exports.wait = (arg) => {
        if (isDebugEval()) {
            if (!fibers.current.delayed)
                fibers.current.delayed = [];
            fibers.current.delayed.push(arg);
            // tslint:disable-next-line:no-string-throw
            throw 'would yield';
        }
        flushDelayed();
        return oldWait(arg);
    };
    const oldRun = exports.run;
    exports.run = (fn) => {
        // tslint:disable-next-line:no-string-throw
        if (isDebugEval())
            throw 'would start a fiber';
        else
            return oldRun(fn);
    };
    console.log('Running with f-promise debugger hooks');
}
//# sourceMappingURL=index.js.map