"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable:no-reference
const chai_1 = require("chai");
const fs = require("fs");
const mzfs = require("mz/fs");
const fsp = require("path");
const __1 = require("..");
const { ok, notOk, equal, notEqual, deepEqual, strictEqual, closeTo, fail, typeOf, isNull, isNotNull, isUndefined, isObject, throws, } = chai_1.assert;
function test(name, fn) {
    it(name, done => {
        __1.run(() => (fn(), undefined)).then(done, done);
    });
}
function delay(val, millis) {
    return __1.wait(cb => {
        setTimeout(() => {
            cb(null, val);
        }, millis || 0);
    });
}
function delayFail(reason, millis) {
    return __1.wait(cb => {
        setTimeout(() => {
            cb(new Error(`reason: ${reason}`));
        }, millis || 0);
    });
}
function sum(array) {
    return array.reduce((sum, x) => sum + (x || 0), 0);
}
process.on('unhandledRejection', err => {
    fail(`Unhandled rejected promise: ${err.stack}`);
});
describe('wait', () => {
    it('promise wait with success', done => {
        const p = __1.run(() => {
            const fname = fsp.join(__dirname, '../../test/f-promise-test.ts');
            const text = __1.wait(mzfs.readFile(fname, 'utf8'));
            typeOf(text, 'string');
            ok(text.length > 200);
            ok(text.indexOf('// tslint') === 0);
            const text2 = __1.wait(mzfs.readFile(fname, 'utf8'));
            equal(text, text2);
            return 'success';
        });
        p.then(result => {
            equal(result, 'success');
            done();
        }, done);
    });
    it('promise wait with error', done => {
        const p = __1.run(() => {
            const fname = fsp.join(__dirname, '../../test/f-promise-test.ts.not.exist');
            __1.wait(mzfs.readFile(fname, 'utf8'));
        });
        p.then(result => {
            fail();
            done();
        }, e => {
            done();
        });
    });
    it('callback wait with success', done => {
        const p = __1.run(() => {
            const fname = fsp.join(__dirname, '../../test/f-promise-test.ts');
            const text = __1.wait(cb => fs.readFile(fname, 'utf8', cb));
            typeOf(text, 'string');
            ok(text.length > 200);
            ok(text.indexOf('// tslint') === 0);
            const text2 = __1.wait(cb => fs.readFile(fname, 'utf8', cb));
            equal(text, text2);
            return 'success';
        });
        p.then(result => {
            equal(result, 'success');
            done();
        }, done);
    });
    it('callback wait with error', done => {
        const p = __1.run(() => {
            const fname = fsp.join(__dirname, '../../test/f-promise-test.ts.not.exist');
            __1.wait(cb => fs.readFile(fname, 'utf8', cb));
        });
        p.then(result => {
            fail();
            done();
        }, e => {
            done();
        });
    });
    it('wait into a callback wait', done => {
        const p = __1.run(() => {
            __1.wait(cb => {
                __1.wait(_cb => process.nextTick(_cb));
                cb(null, null);
            });
        });
        p.then(done, done);
    });
    it('wait throws without run', done => {
        throws(() => __1.wait(_cb => process.nextTick(_cb)), 'cannot wait: no fiber');
        done();
    });
});
describe('queue', () => {
    test('queue overflow', () => {
        const queue = new __1.Queue(2);
        // must produce and consume in parallel to avoid deadlock
        const produce = __1.run(() => {
            queue.write(4);
            queue.write(9);
            queue.write(16);
            queue.write(25);
        });
        const consume = __1.run(() => {
            strictEqual(queue.read(), 4);
            strictEqual(queue.read(), 9);
            strictEqual(queue.read(), 16);
            strictEqual(queue.read(), 25);
        });
        __1.wait(produce);
        __1.wait(consume);
        strictEqual(queue.peek(), undefined);
    });
    test('queue length, contents, alter', () => {
        const queue = new __1.Queue();
        queue.write(4);
        queue.write(9);
        queue.write(16);
        queue.write(25);
        strictEqual(queue.length, 4);
        strictEqual(queue.peek(), 4);
        deepEqual(queue.contents(), [4, 9, 16, 25]);
        queue.adjust(function (arr) {
            return [arr[3], arr[1]];
        });
        strictEqual(queue.peek(), 25);
        strictEqual(queue.read(), 25);
        strictEqual(queue.peek(), 9);
        strictEqual(queue.read(), 9);
        strictEqual(queue.peek(), undefined);
    });
});
describe('handshake', () => {
    test('notify without wait', () => {
        const hk = __1.handshake();
        hk.notify();
        hk.notify();
    });
    test('wait then notify', () => {
        const hk = __1.handshake();
        let counter = 0;
        function runSleepAndCount() {
            __1.run(() => {
                __1.sleep(10);
                counter++;
                hk.notify();
            });
        }
        equal(counter, 0);
        runSleepAndCount();
        hk.wait();
        equal(counter, 1);
        runSleepAndCount();
        hk.wait();
        equal(counter, 2);
    });
    test('multiple wait fails', () => {
        const hk = __1.handshake();
        let thrown = false;
        function runAndWait() {
            __1.run(() => {
                hk.wait();
            }).catch(e => {
                thrown = true;
            });
        }
        runAndWait();
        runAndWait();
        __1.sleep(10);
        hk.notify(); // release not thrown run
        equal(thrown, true);
    });
});
describe('funnel', () => {
    test('less concurrency than allowed', () => {
        const fun = __1.funnel(4);
        const begin = Date.now();
        const results = __1.map([10, 10], timeToSleep => {
            return fun(() => {
                __1.sleep(timeToSleep);
                return 1;
            });
        });
        equal(sum(results), 2);
        closeTo(Date.now() - begin, 10, 4);
    });
    test('more concurrency than allowed', () => {
        const fun = __1.funnel(2);
        const begin = Date.now();
        const results = __1.map([10, 10, 10, 10], timeToSleep => {
            return fun(() => {
                __1.sleep(timeToSleep);
                return 1;
            });
        });
        equal(sum(results), 4);
        closeTo(Date.now() - begin, 20, 8);
    });
    test('close funnel access', () => {
        const fun = __1.funnel(1);
        const begin = Date.now();
        __1.run(() => {
            __1.sleep(15);
            fun.close();
        });
        const results = __1.map([10, 10, 10, 10], timeToSleep => {
            try {
                return fun(() => {
                    __1.sleep(timeToSleep);
                    return 1;
                });
            }
            catch (err) {
                equal(err.message, 'cannot execute: funnel has been closed');
                return 0;
            }
        });
        equal(sum(results), 2);
        closeTo(Date.now() - begin, 20, 8);
    });
});
describe('contexts', () => {
    const mainCx = __1.context();
    it('is main at top level', () => {
        equal(__1.context(), mainCx);
    });
    it('is main inside run', done => {
        __1.run(() => {
            equal(__1.context(), mainCx);
        }).then(done, done);
    });
    it('is scoped inside withContext', done => {
        const cx = {};
        __1.run(() => {
            equal(__1.context(), mainCx);
            __1.withContext(() => {
                equal(__1.context(), cx);
            }, cx);
            equal(__1.context(), mainCx);
        }).then(done, done);
    });
    test('contexts', () => {
        function testContext(x) {
            return __1.withContext(() => {
                const y = delay(2 * x);
                strictEqual(y, 2 * __1.context());
                return y + 1;
            }, x);
        }
        isObject(__1.context());
        const promises = [__1.run(() => testContext(3)), __1.run(() => testContext(5))];
        deepEqual(promises.map(__1.wait), [7, 11]);
        isObject(__1.context());
    });
});
describe('collection functions', () => {
    it('map', done => {
        __1.run(() => {
            deepEqual(__1.map([2, 5], delay), [2, 5]);
            return 'success';
        }).then(result => {
            equal(result, 'success');
            done();
        }, done);
    });
    it('map with error', done => {
        __1.run(() => {
            __1.map([2, 5], delayFail);
            fail();
        })
            .then(_ => fail())
            .catch(err => {
            equal(err.message, 'reason: 2');
            done();
        });
    });
});
describe('canWait', () => {
    it('true inside run', done => {
        __1.run(() => {
            ok(__1.canWait());
            return 'success';
        }).then(result => {
            equal(result, 'success');
            done();
        }, done);
    });
    it('false outside run', () => {
        notOk(__1.canWait());
    });
});
describe('eventHandler', () => {
    it('can wait with it', done => {
        setTimeout(__1.eventHandler(() => {
            ok(__1.canWait());
            done();
        }), 0);
    });
    it('cannot wait without', done => {
        setTimeout(() => {
            notOk(__1.canWait());
            done();
        }, 0);
    });
    it('outside run', done => {
        notOk(__1.canWait());
        let sync = true;
        __1.eventHandler((arg) => {
            equal(arg, 'hello', 'arg ok');
            __1.wait(cb => setTimeout(cb, 0));
            equal(sync, false, 'new fiber');
            done();
        })('hello');
        sync = false;
    });
    it('inside run', done => {
        __1.run(() => {
            let sync = true;
            ok(__1.canWait());
            __1.eventHandler((arg) => {
                equal(arg, 'hello', 'arg ok');
                __1.wait(cb => setTimeout(cb, 0));
                equal(sync, true, 'same fiber as run');
                done();
            })('hello');
            sync = false;
        });
    });
    it('preserves arity', () => {
        equal(__1.eventHandler(() => { }).length, 0);
        equal(__1.eventHandler((a, b) => { }).length, 2);
    });
    it('starts with a fresh context if outside run', done => {
        ok(!__1.canWait());
        __1.eventHandler(() => {
            isNotNull(__1.context());
            done();
        })();
    });
    it('preserves context if already inside run', done => {
        __1.run(() => {
            ok(__1.canWait());
            const cx = {};
            __1.withContext(() => {
                __1.eventHandler(() => {
                    equal(__1.context(), cx);
                    done();
                })();
            }, cx);
        });
    });
});
//# sourceMappingURL=f-promise-test.js.map